; video.asm - Custom ZX81 Video Driver. Paul Daniels

; 21/04/2010 - PD Initial version
;                       Mark frame as blank (all border)
;                       ISR available after top border
;                       ISR available after bottom border
;                       Set 1st and last line of D_FILE
;                       Handle two D_FILES, one per frame
;                       ISR available during VSYNC *1

; D_SETFR - Set frames to display.

; D_TWAIT - Wait for start of top margin.
; D_TISR - Set top margin ISR.

; D_BWAIT - Wait for start of bottom margin.
; D_BISR - Set bottom margin ISR.

; D_F0ADDR - Set address of D_FILE0.
; D_F0WAIT - Returns when/if frame 1 is being displayed.
; D_F0SET  - Set offset and length of frame 1.

; D_F1ADDR - Set address of D_FILE1.
; D_F1WAIT - Returns when/if frame 2 is being displayed.
; D_F1SET  - Set offset and length of frame 2.

; *1 ISR has to be 1976 cycles long.
; *2 Keyboard scan to be added by user.

;*******************************************************************************
; The ZX81 video driver is tightly connected to how a TV screen is displayed.
;
; The driver has to provide a VSYNC pulse and react to HSYNC pulses.
;
; A TV FRAME is made up as follows:
;
;       Name    Time    Rasters Desc
;       VSYNC   400us   6.25    Time for election beam to move to top of screen
;       TOP     3520us  55      Rasters to center MAIN display
;               64us    1       Change over from TOP to MAIN display
;       MAIN    12288us 192     MAIN display takes 192 rasters
;       BOTTOM  3520us  55      Rasters to complete frame
;       ======= 19792us 309     Ideal would be 20000us / 312.5 rasters.

; TOP and BOTTOM borders are handled by NMI interrupts.
; MAIN display is handled by IM1 interrupts.

; IM1 routine at $0038 doesn't use IY or system variables.
; NMI routine at $0066 doesn't use IY. Uses $400C (D_FILE).

; NB: VSYNC in ROM might take longer then 400us, 608us would be ideal.

; ZX81 runs at 3.25Mhz. This gives 207 cycles per raster line.

; NMI takes 44 cycles, leaving 163 cycles for main program = 207 cycles
; IM1 takes 71 cycles. Outputting 34 chars takes 136 cycles = 207 cycles

; 1 cycle = 0.3076923us. VSYNC should be 1976 cycles.

; VSYNC created with "IN A, ($FE) / wait >400us / OUT ($FF), A".

; Reset LCNTR in ULA with "IN A, ($FE) / OUT ($FF), A" with NMI int off.

;*******************************************************************************
; D_START       Change over to custom video driver.
;
; Entry		None
; Exit		Destroys A, HL. IX set to driver addr.
;*******************************************************************************
D_START
		ld a, (MARGIN)          ; Put MARGIN out of the way.
		ld (D_MARG), a
		ld hl, (D_FILE)         ; Put D_FILE out of the way.
		ld (D_FILE0), hl
		ld (D_FILE1), hl

		ld hl, $0018            ; Display 0 offset, 24 lines
		call D_F0SET
		call D_F1SET

		; Busy wait until FRAMES sysvar is changed, signifying VSYNC.

		ld hl, FRAMES
		ld a, (hl)
-               cp (hl)
		jr z, -

		; Rom will now be handling top border.
		; Set ix to the routine to display main screen.

		ld ix, D_SCREEN

		ret

;*******************************************************************************
; D_SETFR       Set frames to display. Reg A set with bit 0 and/or 1 set.
;
; Entry		A set to frames to display.
; Exit		No registers affected.
;*******************************************************************************
D_SETFR
		and $03
		ld (D_FLAGS), a
		ret

;*******************************************************************************
; D_TWAIT       Wait for start of top border.
;
; Entry		None.
; Exit		No registers affected.
;*******************************************************************************
D_TWAIT
		push hl
		ld hl, D_FLAGS
		res 2, (hl)
-		bit 2, (hl)
		jr z, -
		pop hl
		ret

;*******************************************************************************
; D_BWAIT       Wait for start of bottom border.
;
; Entry		None.
; Exit		No registers affected.
;*******************************************************************************
D_BWAIT
		push hl
		ld hl, D_FLAGS
		res 3, (hl)
-		bit 3, (hl)
		jr z, -
		pop hl
		ret

;*******************************************************************************
; D_F0WAIT	Wait for frame 0.
;
; Entry		None.
; Exit		No registers affected.
;*******************************************************************************
D_F0WAIT
		push hl
		ld hl, D_FRAME
-		bit 0, (hl)
		jr z, -
		pop hl
		ret

;*******************************************************************************
; D_F1WAIT	Wait for frame 1.
;
; Entry		None.
; Exit		No registers affected.
;*******************************************************************************
D_F1WAIT
		push hl
		ld hl, D_FRAME
-		bit 0, (hl)
		jr nz, -
		pop hl
		ret

;*******************************************************************************
; D_F0ADDR	Set address of D_FILE0 frame
;
; Entry		HL set to DFILE address.
; Exit		No registers affected.
;*******************************************************************************
D_F0ADDR
		ld (D_FILE0), hl
		ret

;*******************************************************************************
; D_F1ADDR	Set address of D_FILE1 frame
;
; Entry		HL set to DFILE address.
; Exit		No registers affected.
;*******************************************************************************
D_F1ADDR
		ld (D_FILE1), hl
		ret

;*******************************************************************************
; D_TISR        Set address of top ISR
;
; Entry		HL set to top margin ISR address.
; Exit		No registers affected.
;*******************************************************************************
D_TISR
		ld (D_TOP1+1), hl
		ret

;*******************************************************************************
; D_BISR        Set address of bottom ISR
;
; Entry		HL set to bottom margin ISR address.
; Exit		No registers affected.
;*******************************************************************************
D_BISR
		ld (D_BOT1+1), hl
		ret

;*******************************************************************************
; D_F0SET       Set frame 0 char line offset and length
;
; Entry		H set to offset 0 - 23. L set to length 0 - 24.
; Exit		No registers affected.
;*******************************************************************************
D_F0SET
		push af
		push bc

		ld a, h                 ; offset
		sla a
		sla a
		sla a
		ld c, a
		ld a, (D_MARG)
		add a, c
		neg
		inc a
		ld (D_F0MTOP), a	; Margin Top Temp

		ld a, l                 ; length
		sla a
		sla a
		sla a
		ld b, a
		ld a, 192
		sub b
		sub c
		ld c, a
		ld a, (D_MARG)
		add a, c
		neg
		inc a
		ld (D_F0MBOT), a        ; Margin Bottom Temp

		ld a, l
		inc a
		ld (D_F0LEN), a

		pop bc
		pop af
		ret
		
;*******************************************************************************
; D_F1SET       Set frame 1 char line offset and length
;
; Entry		H set to offset 0 - 23. L set to length 0 - 24.
; Exit		No registers affected.
;*******************************************************************************
D_F1SET
		push af
		push bc

		ld a, h                 ; offset
		sla a
		sla a
		sla a
		ld c, a
		ld a, (D_MARG)
		add a, c
		neg
		inc a
		ld (D_F1MTOP), a	; Margin Top Temp

		ld a, l                 ; length
		sla a
		sla a
		sla a
		ld b, a
		ld a, 192
		sub b
		sub c
		ld c, a
		ld a, (D_MARG)
		add a, c
		neg
		inc a
		ld (D_F1MBOT), a        ; Margin Bottom Temp

		ld a, l
		inc a
		ld (D_F1LEN), a

		pop bc
		pop af
		ret
		
;*******************************************************************************
; D_VSYNC       Generate VSYNC
;*******************************************************************************
D_VSYNC
		in a, ($fe)             ; [10]

		ld b, 9                 ; [7]

		; Delay of 206 cycles. Close enough to 207 for now.

--		ld c, 12                ; [7]
-		dec c                   ; [4]
		jr nz, -                ; [(12*loop-1)+7]

-		djnz --                 ; [(13*loop-1)+8]

		out ($ff), a            ; [11]

;*******************************************************************************
; D_TOP         Setup driver for top margin.
;*******************************************************************************
D_TOP
		ld a, (D_MTOP)
		ex af, af'
		out ($fe),a
		ld ix, D_SCREEN

		; Check if frame is marked as blank
		; Update D_FILE.

		ld hl, (D_FILE0)                        ; Update DFILE
		ld a, (D_FRAME)
		xor 1
		ld (D_FRAME), a
		jr nz, +                                ; Update DFILE
		ld hl, (D_FILE1)                        ; Update DFILE
+		ld (D_FILE), hl                         ; Update DFILE

		ld hl, D_FLAGS
		inc a
		ld c,a
		ld a, (hl)
		and $03
		and c
		jr nz, +
		ld ix, D_BLANK
+
		; Set flag for top border
		
		set 2, (hl)

		; Check for users ISR

		ld a, (D_TOP1+2)
		and a
		jr z, +
D_TOP1		call 0
+

		pop hl
		pop de
		pop bc
		pop af

		ret

;*******************************************************************************
; D_BLANK       Output blank frame
;*******************************************************************************
D_BLANK
D_BLN1		ld a, 193
		neg
		inc a
		ex af, af'
		out ($fe),a
		ld ix, D_BOTTOM

		; Check for users ISR?

		pop hl
		pop de
		pop bc
		pop af

		ret

;*******************************************************************************
; D_SCREEN      Output display frame
;*******************************************************************************
D_SCREEN

		; Allow for variable number of lines

		ld a,r
D_SCN1		ld bc, $1901                    ; Modified
		ld a, $f5
		call $02b5

;*******************************************************************************
; D_BOTTOM      Setup driver for bottom margin.
;*******************************************************************************
D_BOTTOM
		ld a, (D_MBOT)
		ex af, af'
		out ($fe),a
		ld ix, D_VSYNC
		ld hl, D_FLAGS

		; Set flag for bottom border
		
		set 3, (hl)

		; Update top and bottom border sizes

		ld a, (D_FRAME)
		and a
		jr nz, D_BOT3

		ld a, (D_F0LEN)
		ld (D_SCN1+2), a
		sla a
		sla a
		sla a
		inc a
		ld (D_BLN1+1), a

		ld a, (D_F0MTOP)
		ld (D_MTOP), a
		ld a, (D_F0MBOT)
		ld (D_MBOT), a
		jr D_BOT2
D_BOT3
		ld a, (D_F1LEN)
		ld (D_SCN1+2), a
		sla a
		sla a
		sla a
		inc a
		ld (D_BLN1+1), a

		ld a, (D_F1MTOP)
		ld (D_MTOP), a
		ld a, (D_F1MBOT)
		ld (D_MBOT), a
D_BOT2
		; Check for users ISR

		ld a, (D_BOT1+2)
		and a
		jr z, +
D_BOT1		call 0
+
		pop hl
		pop de
		pop bc
		pop af

		ret

;*******************************************************************************
; VARIABLES
;*******************************************************************************

D_MARG		.DB 0		; Copy of MARGIN

D_FLAGS		.DB $03		; Bit	Description
				; 0	Display frame 0 when set
				; 1	Display frame 1 when set

D_FRAME		.DB 0		; Keeps track of frame being displayed, 0 or 1.
D_MTOP		.DB 0		; Total of rasters in top margin
D_MBOT		.DB 0		; Total of rasters in bottom margin
D_FILE0		.DW 0		; Address of frame 0 DFILE
D_FILE1		.DW 0		; Address of frame 1 DFILE
D_F0LEN		.DB 0		; Temp frame 0 length
D_F0MTOP	.DB 0		; Temp frame 0 top margin
D_F0MBOT	.DB 0		; Temp frame 0 bottom margin
D_F1LEN		.DB 0		; Temp frame 1 length
D_F1MTOP	.DB 0		; Temp frame 1 top margin
D_F1MBOT	.DB 0		; Temp frame 1 bottom margin
